#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file tkviewer.py
 @brief This module provides the graphical user interface
 @date $Date$

 @author 株式会社 東日本計算センター

 プロジェクト方針決定後に記載

"""
import sys
import time
import RTC
import WiFiNodeIF
import Tkinter as Tk
import ttk
from logging import getLogger, StreamHandler
from logging import INFO as LOG_LV

sys.path.append(".")

# Setup logger
logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(LOG_LV)
handler.setLevel(LOG_LV)
logger.addHandler(handler)

# Set constants
OBSERVER_NODE = 1  # 0b001
NEIGHBOR_NODE = 2  # 0b010
STATUS = 4  # 0b100
ALL = OBSERVER_NODE | NEIGHBOR_NODE | STATUS  # 0b111


##
# @class NeighborNodeList
# @brief ロボット(Observer)一覧を表示するフレーム(ツリー)
#
class ObserverNodeList(Tk.Frame):

    def __init__(self, master=None):
        Tk.Frame.__init__(self, master)

        self._treeview = None
        self._sel_id = None
        self._plugin_func = None  # 用途不明
        self._is_descending = False

        # configure observer_node list
        self._treeview = ttk.Treeview(self, columns="ObserverNode",
                                      show="headings", selectmode="browse")

        cb_func = (lambda col="ObserverNode":
                   self._cb_sort_order_change(col, not self._is_descending))

        self._treeview.heading(
            "ObserverNode", text="ObserverNode", anchor=Tk.CENTER, command=cb_func)

        self._treeview.column("ObserverNode", stretch=True, anchor=Tk.CENTER,
                              minwidth=180, width=180)
        self._treeview.bind(
            "<<TreeviewSelect>>", self._cb_sel_observer_node_change)

        # configure scrollbar
        sb = ttk.Scrollbar(
            self, orient=Tk.VERTICAL, command=self._treeview.yview)
        self._treeview.configure(yscroll=sb.set)

        # add configured widgets
        self._treeview.grid(row=0, column=0, sticky=Tk.NSEW)
        sb.grid(row=0, column=1, sticky=Tk.NS)

        # set resize properties
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

    ##
    # @brief
    def add(self, observer_node_id):
        self._treeview.insert(
            "", "end", value=observer_node_id, tags=observer_node_id)
        self._treeview.tag_configure(observer_node_id, foreground="black")

        self._cb_sort_order_change("ObserverNode", self._is_descending)

    ##
    # @brif もっと単純にできないか? itemオブジェクトについて調べる
    def clear(self):
        observer_node_list = self._treeview.get_children()
        for i in observer_node_list:
            self._treeview.delete(i)

    ##
    # @brief Oberverからのデータが途絶えた時の処理。ノードIDの表示色を赤に変更する。
    def set_missing_observer_node(self, observer_node_id, is_missing):
        child = self._treeview.get_children()

        target = None
        for item in child:
            if observer_node_id == str(self._treeview.item(item)['values'][0]):
                target = item
                break

        if target:
            color = "black"
            if is_missing:
                color = "red"

            self._treeview.tag_configure(observer_node_id, foreground=color)

    ##
    # @brief 表示中のObserverノード数
    def get_observer_node_count(self):
        cnt = len(self._treeview.get_children())
        return cnt

    def get_sel_observer_node_id(self):
        return self._sel_id

    def set_sel_observer_node_id(self, observer_node_id):
        child = self._treeview.get_children()

        target = None
        for item in child:
            if observer_node_id == str(self._treeview.item(item)['values'][0]):
                target = item
                break

        if target:
            self._treeview.selection_set(target)

    ##
    # @brief 
    def plugin_sel_observer_node_change(self, func):
        self._plugin_func = func

    ##
    # @brief ttk.Treeview <<TreeviewSelect>>イベントハンドラ
    def _cb_sel_observer_node_change(self, event):
        sel_item = self._treeview.selection()

        if not sel_item:
            return

        child = self._treeview.get_children()

        for item in child:
            if sel_item[0] == item:
                self._sel_id = str(self._treeview.item(item)['values'][0])
                break

        # Callback
        if self._plugin_func:
            self._plugin_func()

    def _cb_sort_order_change(self, col, sort_order):
        child = [(self._treeview.set(child, col), child)
                 for child in self._treeview.get_children()]

        child.sort(reverse=sort_order)

        for i, item in enumerate(child):
            self._treeview.move(item[1], "", i)

        self._is_descending = sort_order

##
# @class NeighborNodeList
# @brief 近隣ノードを表示するフレーム(ツリー)。
#


class NeighborNodeList(Tk.Frame):

    def __init__(self, master=None):
        Tk.Frame.__init__(self, master)

        self._treeview = None
        self._is_descending = False

        # configure neighbor_node list
        col_name = ("MAC Address", "Channel", "Signal Level", "ESSID")
        self._treeview = ttk.Treeview(self, columns=col_name, show="headings")

        for name in col_name:
            cb_func = (lambda col=name:
                       self._cb_sort_order_change(col, not self._is_descending))

            self._treeview.heading(name, text=name.title(),
                                   anchor=Tk.CENTER,
                                   command=cb_func)

            self._treeview.column(name, stretch=True, anchor=Tk.CENTER,
                                  minwidth=150, width=150)

        # configure scrollbar
        sb = ttk.Scrollbar(
            self, orient=Tk.VERTICAL, command=self._treeview.yview)
        self._treeview.configure(yscroll=sb.set)

        # add configured widgets
        self._treeview.grid(row=0, column=0, sticky=Tk.NSEW)
        sb.grid(row=0, column=1, sticky=Tk.NS)

        # set resize properties
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

    def add(self, mac_addres, channel, signal_level, essid, tagname=None):
        self._treeview.insert("", "end",
                              values=(
                                  mac_addres, channel, signal_level, essid),
                              tags="neighbor_node")

    def clear(self):
        child = self._treeview.get_children()
        for i in child:
            self._treeview.delete(i)

    ##
    # @brief 再描画。
    #
    # @param self
    # @param neighbor_node
    #
    def show_entire_neighbor_node(self, neighbor_node):
        self.clear()

        if neighbor_node:
            for i in neighbor_node:
                self.add(i.mac_address, i.channel, i.signal_level, i.essid,)

            self._treeview.tag_configure("neighbor_node", foreground="black")
        else:
            self.add("no data", "no data", "no data", "no data")
            self._treeview.tag_configure("neighbor_node", foreground="red")

    ##
    # @brief
    def _cb_sort_order_change(self, col, sort_order):
        child = [(self._treeview.set(child, col), child)
                 for child in self._treeview.get_children()]

        child.sort(reverse=sort_order)

        for i, item in enumerate(child):
            self._treeview.move(item[1], "", i)

        self._is_descending = sort_order


##
# @class StatusBar
# @brief ステータスバー。
#
class StatusBar(Tk.Frame):

    def __init__(self, master=None):
        Tk.Frame.__init__(self, master)
        self._label = None

        self._label = Tk.Label(self, bd=1, relief=Tk.SUNKEN, anchor=Tk.E)
        self._label.pack(fill=Tk.X)

    def set(self, format, *args):
        color = "black"

        if not args:
            color = "red"

        self._label.config(text=format % args, fg=color)
        self._label.update_idletasks()

    def clear(self):
        self._label.config(text="")
        self._label.update_idletasks()


##
# @class TkWiFiNeighborNodeViewer
# @brief メインウィンドウ。
#
class TkWiFiNeighborNodeViewer(Tk.Frame):

    def __init__(self, master=None):
        Tk.Frame.__init__(self, master)
        self._observer_node_list = None
        self._neighbor_node_list = None
        self._status_bar = None
        self._data_src_func = None  # Observer からの受信データ

        self.master.title("WiFi Neighbor Node Viewer")

        # init sub-frame
        self._observer_node_list = ObserverNodeList()
        self._neighbor_node_list = NeighborNodeList()
        self._status_bar = StatusBar()

        # イベントハンドラ登録 (ノード名がクリックされたら、再描画する)
        self._observer_node_list.plugin_sel_observer_node_change(
            self._cb_selected_observer_node_change)

        # add sub-frame
        self._observer_node_list.grid(row=0, column=0, sticky=Tk.NSEW)
        self._neighbor_node_list.grid(row=0, column=1, sticky=Tk.NSEW)
        self._status_bar.grid(columnspan=2, row=1, column=0, sticky=Tk.EW)

        # set resize properties
        self.master.grid_rowconfigure(0, weight=1)
        self.master.grid_columnconfigure(0, weight=1)
        self.master.grid_columnconfigure(1, weight=3)

        # set ttk style
        ttk.Style().map("Treeview",
                        background=[("selected", "LightSkyBlue")],
                        foreground=[])

    ##
    # @brief Setter。イベントハンドラを登録する。
    def set_data_src_func(self, func):
        self._data_src_func = func

    def init_viewer(self):
        self._observer_node_list.clear()
        self._neighbor_node_list.clear()
        self._status_bar.clear()

    ##
    # @brief 。
    #
    # @param target 更新対象ペイン
    # @param observer_node_id オブザーバーノードID
    # @param **kwargs オブザーバーノードの辞書
    def draw_gui(self, target, observer_node_id, **kwargs):
        # 論理演算(フラグ)にて更新するサブフレームを判定している
        if target & OBSERVER_NODE:
            self._draw_observer_node_list(observer_node_id, kwargs)

        sel_idx = self._observer_node_list.get_sel_observer_node_id()
        if sel_idx is None:
            return

        if sel_idx != observer_node_id:
            return

        # Observer からの受信データをセット
        neighbor_node_list = self._data_src_func(observer_node_id)

        if target & NEIGHBOR_NODE:
            self._draw_neighbor_node_list(neighbor_node_list)

        if target & STATUS:
            self._draw_status_bar(neighbor_node_list)

    ##
    # @brief オブザーバーノード一覧を再描画する
    # @param observer_node_id
    # @param kwargs
    def _draw_observer_node_list(self, observer_node_id, kwargs):
        is_missing = None
        try:
            is_missing = kwargs["missing"]
        except KeyError:
            pass

        if is_missing is None:
            # add observer_node
            old_cnt = self._observer_node_list.get_observer_node_count()

            self._observer_node_list.add(observer_node_id)

            if not old_cnt:
                self._observer_node_list.set_sel_observer_node_id(
                    observer_node_id)
        else:
            # update missing observer_node
            self._observer_node_list.set_missing_observer_node(
                observer_node_id, is_missing)

    ##
    # @brief 近隣ノード一覧を再描画する
    # @param neighbor_node_list
    def _draw_neighbor_node_list(self, neighbor_node_list):
        if not neighbor_node_list:
            return

        self._neighbor_node_list.show_entire_neighbor_node(
            neighbor_node_list.data)

    ##
    # @brief ステータスバーを再描画する
    # @param neighbor_node_list Observerからの受信データ
    def _draw_status_bar(self, neighbor_node_list):
        if not neighbor_node_list:
            return

        if neighbor_node_list.data:
            tm = time.localtime(neighbor_node_list.tm.sec)
            self._status_bar.set("updated on %04d/%02d/%02d %02d:%02d:%02d",
                                 tm.tm_year, tm.tm_mon, tm.tm_mday,
                                 tm.tm_hour, tm.tm_min, tm.tm_sec)
        else:
            info_str = ("Communication route to this observer_node"
                        " might have been disconnected")
            self._status_bar.set(info_str)

    ##
    # @brief ノード名がクリックされたら、近隣ノード一覧とステータスバーを再描画する
    def _cb_selected_observer_node_change(self):
        sel_id = self._observer_node_list.get_sel_observer_node_id()
        self.draw_gui(NEIGHBOR_NODE | STATUS, sel_id)


##
# @class dummyViewer
# @brief デバッグ用クラス。
#
class dummyViewer(object):

    def __init__(self):
        self._neighbor_node_data = []

        self.make_test_data()

    def get_neighbor_node_data(self, observer_node_id):
        ret = None

        for data in self._neighbor_node_data:
            if observer_node_id == data.id:
                ret = data
                break

        if not ret:
            logger.error("invalid requested idx: %d", idx)

        return ret

    def make_test_data(self):
        for i in range(1, 31):
            tmp_observer_node = WiFiNodeIF.TimedNeighborNodeList(
                RTC.Time(0, 0), 0, [])

            for j in range(1, 51):
                """
                data = "192.168.%03d.%03d" % (i, j)
                tmp_neighbor_node = WiFiNodeIF.NeighborNode(data, data, data, data)
                """
                tmp_neighbor_node = WiFiNodeIF.NeighborNode(self.gen_random_str(5),
                                                            self.gen_random_str(
                                                                5),
                                                            self.gen_random_str(
                                                                5),
                                                            self.gen_random_str(5))

                tmp_observer_node.data.append(tmp_neighbor_node)

            tmp_observer_node.tm.sec = 1595588400
            #tmp_observer_node.id = "drone_%02d" % (i)
            tmp_observer_node.id = self.gen_random_str(5)

            self._neighbor_node_data.append(tmp_observer_node)

    def clear_test_data(self, idx):
        if idx < len(self._neighbor_node_data):
            del self._neighbor_node_data[idx].data[:]

    def gen_random_str(self, len):
        import string
        import random

        src = string.digits + string.letters
        ret = ''
        for i in range(1, len):
            ret += random.choice(src)

        return ret


##
# @brief デバッグ用メソッド。
#
# デバッグ用メソッド。
# __name__ == "__main__" の場合に自動実行される。
def debug_tk_viewer():
    dummy = dummyViewer()
    tkv = TkWiFiNeighborNodeViewer()
    tkv.set_data_src_func(dummy.get_neighbor_node_data)

    for i in dummy._neighbor_node_data:
        tkv.draw_gui(OBSERVER_NODE, i.id)

    dummy.clear_test_data(1)
    tkv.draw_gui(ALL, dummy._neighbor_node_data[1].id, missing=True)
    dummy.clear_test_data(3)
    tkv.draw_gui(ALL, dummy._neighbor_node_data[3].id, missing=True)
    dummy.clear_test_data(5)
    tkv.draw_gui(ALL, dummy._neighbor_node_data[5].id, missing=True)

    tkv.mainloop()

if __name__ == "__main__":
    debug_tk_viewer()
